"""
Main patcher utility
"""

from pathlib import Path
from re import sub
from typing import Dict, Optional, Union, cast

from archinfo import Arch

from pypatches.binary_manager import BinaryManager
from pypatches.patches import (
    AddCodePatch,
    AlwaysBranchPatch,
    CallerReplacePatch,
    DataPatch,
    FiniPatch,
    FunctionReplacePatch,
    InitPatch,
    InvertBranchPatch,
    NeverBranchPatch,
    NopPatch,
    PatchType,
    ReplaceCodePatch,
    SkipAndReturnPatch,
)


class Patcher:
    """Patch processor, applies patches to the target binary

    Args:
        binary: Either a path to the target binary or the raw bytes
            of the binary
        cle_opts: An optional replacement set of options to pass to
            cle.Loader
    """

    def __init__(
        self,
        binary: Union[Path, str, bytes],
        cle_opts: Optional[Dict[str, bool]] = None,
    ) -> None:
        """Set up patcher with the target binary"""
        self.binary = BinaryManager(binary, cle_opts)

    def apply(self, patch: PatchType) -> None:
        """Apply a patch to the target binary

        Args:
            patch: The patch to apply, can be any of the patch types
        """

        def camel_to_snake(name: str) -> str:
            """
            Convert a camel case string to snake case
            """
            name = sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
            return sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

        patch_dispatch_func_name = f"apply_{camel_to_snake(patch.__class__.__name__)}"
        if not hasattr(self, patch_dispatch_func_name):
            raise NotImplementedError(
                f"Patch type {patch.__class__.__name__} is not supported."
            )

        patch_dispatch_func = getattr(self, patch_dispatch_func_name)
        patch_dispatch_func(patch)

    def save(self, path: Union[str, Path]) -> None:
        """
        Save the patched binary to the given path
        """
        if isinstance(path, str):
            path = Path(path)

        self.binary.save(path)

        path.chmod(0o755)

    def apply_nop_patch(self, patch: NopPatch) -> None:
        """
        Apply a nop patch to the target binary
        """
        nop_len = len(cast(Arch, self.binary.cle_binary.arch).nop_instruction)

        for address_range in patch.address_ranges:
            for address in range(
                address_range.start,
                max(address_range.end, address_range.start + nop_len),
                nop_len,
            ):
                self.binary.write(
                    address, cast(Arch, self.binary.cle_binary.arch).nop_instruction
                )

    def apply_invert_branch_patch(self, patch: InvertBranchPatch) -> None:
        """
        Apply a branch patch to the target binary
        """
        raise NotImplementedError("Branch patches are not supported.")

    def apply_always_branch_patch(self, patch: AlwaysBranchPatch) -> None:
        """
        Apply a branch patch to the target binary
        """

    def apply_never_branch_patch(self, patch: NeverBranchPatch) -> None:
        """
        Apply a branch patch to the target binary
        """
        raise NotImplementedError("Branch patches are not supported.")

    def apply_skip_and_return_patch(self, patch: SkipAndReturnPatch) -> None:
        """
        Apply a branch patch to the target binary
        """
        raise NotImplementedError("Branch patches are not supported.")

    def apply_function_replace_patch(self, patch: FunctionReplacePatch) -> None:
        """
        Apply a function replace patch to the target binary
        """
        raise NotImplementedError("Function replace patches are not supported.")

    def apply_caller_replace_patch(self, patch: CallerReplacePatch) -> None:
        """
        Apply a caller replace patch to the target binary
        """
        raise NotImplementedError("Caller replace patches are not supported.")

    def apply_init_patch(self, patch: InitPatch) -> None:
        """
        Apply an init patch to the target binary
        """
        raise NotImplementedError("Init patches are not supported.")

    def apply_fini_patch(self, patch: FiniPatch) -> None:
        """
        Apply a fini patch to the target binary
        """
        raise NotImplementedError("Fini patches are not supported.")

    def apply_data_patch(self, patch: DataPatch) -> None:
        """
        Apply a data patch to the target binary
        """
        self.binary.add_data(patch.data, patch.label)

    def apply_add_code_patch(self, patch: AddCodePatch) -> None:
        """
        Apply an add code patch to the target binary
        """
        self.binary.add_code(patch.code, patch.label)

    def apply_replace_code_patch(self, patch: ReplaceCodePatch) -> None:
        """
        Apply a replace code patch to the target binary
        """
        self.binary.write(patch.address, patch.code)



"""Reexports of the code module.

This module reexports the code module to make it easier to import the code
from external modules.

"""

from pypatches.code.code import Code
from pypatches.code.c_code import CCode, build_c_code
from pypatches.code.asm import ASMCode
from pypatches.code.raw import RawCode



"""Binary program wrapper providing modification and analysis via LIEF and angr
"""

from dataclasses import dataclass
from io import BytesIO
from logging import getLogger
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Dict, List, Optional, Union, cast

from angr import Project
from cle.backends import Backend
from lief import Binary as LIEFBinary  # pylint: disable=no-name-in-module
from lief import parse  # pylint: disable=no-name-in-module
from lief.ELF import (  # pylint: disable=no-name-in-module,import-error
    SEGMENT_FLAGS,
    SEGMENT_TYPES,
    Segment,
)
from capstone import CsInsn

from pypatches.error import NoSectionError
from pypatches.code.code import Code
from pypatches.transform.info import TransformInfo
from pypatches.write_operation import WriteOperation

logger = getLogger(__name__)


class BinaryManager:
    """Wrapper for a binary program to provide information for patching

    Attributes:
        path: The path to the binary, if it exists on disk
        blob: The binary blob of the binary
        lief_binary: The LIEF binary object, which provides the majority of the program
            information for modification
        angr_project: The angr project, which provides the CFG and other information
            for analysis and deep inspection
        cle_binary: The CLE binary object, which provides angr's loader information
        cle_opts: The kwargs passed to `cle.Loader`, which can be used to override
            the default loader options
        cfg_opts: The kwargs passed to `angr.Project.analyses.CFGFast`, which can be
            used to override the default CFG options
        writes: A list of `WriteOperation`s that are queued for application to the
            binary
        code_to_add: A dictionary of `Code` objects to add to the binary, keyed by
            the label they are associated with
        alignment: The alignment to use when aligning addresses
        data_to_add: A dictionary of data to add to the binary, keyed by the label
            the data is associated with

    Args:
        binary: If `binary` is a `Path` or `str` we will attempt to load it from that
            path on disk. If it is `bytes`, we will load the bytes as a binary blob.
        cle_opts: If provided, will override the kwargs passed to `cle.Loader`, defaults
            to sane options.
        cfg_opts: If provided, will override the kwargs passed to
            `angr.Project.analyses.CFGFast`, defaults to sane options.
        silence_angr_logs: If `True`, will silence angr's logging output, defaults to
            `False`
        alignment: The alignment to use when aligning addresses, defaults to 8, but will
            be overridden by lief's alignment if it discovers a different alignment when
            loading the binary

    """

    path: Optional[Path] = None
    blob: BytesIO
    lief_binary: LIEFBinary
    angr_project: Project
    cle_binary: Backend
    cle_opts: Dict[str, bool] = {
        "auto_load_libs": False,
        "use_system_libs": False,
    }
    cfg_opts: Dict[str, bool] = {
        "normalize": True,
        "data_references": False,
        "cross_references": False,
        "skip_unmapped_addrs": True,
        "force_complete_scan": False,
    }
    writes: List[WriteOperation] = []
    code_to_add: Dict[str, Code] = {}
    alignment: int = 0x1000
    data_to_add: Dict[str, bytes] = {}

    def __init__(
        self,
        binary: Union[Path, str, bytes],
        cle_opts: Optional[Dict[str, bool]] = None,
        cfg_opts: Optional[Dict[str, bool]] = None,
        silence_angr_logs: bool = False,
        alignment: int = 8,
    ) -> None:
        """Initialize the binary wrapper via one of several methods."""
        if isinstance(binary, Path):
            self.path = binary
        elif isinstance(binary, str):
            self.path = Path(binary)
        elif isinstance(binary, bytes):
            self.path = None
        else:
            raise TypeError(
                f"Requested binary is of type {type(binary)}"
                ", expected Path, str, or bytes."
            )

        if self.path is not None and not self.path.is_file():
            raise FileNotFoundError(
                f"Requested binary {self.path} was not found or could not be opened."
            )

        if cle_opts is not None:
            self.cle_opts = cle_opts

        if cfg_opts is not None:
            self.cfg_opts = cfg_opts

        if silence_angr_logs:
            for logger_name in ("angr", "pyvex", "cle", "archinfo", "claripy"):
                getLogger(logger_name).setLevel("ERROR")

        self.alignment = alignment

        if self.path is None:
            self.blob = BytesIO(cast(bytes, binary))
        else:
            self.blob = BytesIO(self.path.read_bytes())

        self.load_lief_binary()
        self.load_angr_project()

    def reload_blob_from_lief(self) -> None:
        """Reload the blob from the current LIEF binary"""
        tempfile = NamedTemporaryFile(delete=False)
        temppath = Path(tempfile.name)

        try:
            tempfile.close()
            self.lief_binary.write(str(temppath.resolve()))
            self.blob = BytesIO(temppath.read_bytes())
        except Exception as e:
            raise e
        finally:
            temppath.unlink(missing_ok=True)

        self.load_lief_binary()
        self.load_angr_project()

    def load_lief_binary(self) -> None:
        """Load the binary into LIEF"""
        self.lief_binary = parse(self.blob.getbuffer())

    def load_angr_project(self) -> None:
        """Load the angr project from the binary blob"""
        self.angr_project = Project(
            self.blob,
            main_opts={"base_addr": self.lief_binary.imagebase},
            load_options=self.cle_opts,
        )
        self.angr_project.analyses.CFGFast(  # type: ignore
            **self.cfg_opts,
        )

        if self.angr_project.loader.main_object is None:
            raise FileNotFoundError(
                f"Requested binary {self.path} " "was not found or could not be opened."
            )

        self.cle_binary = self.angr_project.loader.main_object

    def align(
        self,
        address: int,
        alignment: Optional[int] = None,
    ) -> int:
        """Align an address to the specified alignment. If no alignment is
        provided, the default alignment will be used.

        Args:
            address: The address to align
            alignment: The alignment to use, defaults to the default alignment of the
                binary
        """
        if alignment is None:
            alignment = self.alignment

        address = (address + (alignment - 1)) & ~(alignment - 1)

        return address

    def write(
        self,
        where: Union[
            str,
            Callable[[TransformInfo], int],
            int,
            List[int],
            Callable[[TransformInfo], List[int]],
        ],
        data: Union[bytes, Callable[[Dict[str, int]], bytes], Code],
    ) -> None:
        """Write data to the binary at the given virtual address

        Args:

            where: Either an address or a label that will resolve to an address, or
                a function that takes the transform info after segment modification and
                returns an address
            data: The bytes or a function that takes a dictionary of labels
                and addresses and returns bytes to write to the patch
        """

        operation = WriteOperation(data, where)
        logger.info(f"Queuing write operation: {operation}")
        self.writes.append(operation)

    def read(self, vaddr: int, size: int) -> bytes:
        """Read data from the binary at the given virtual address

        Args:
            vaddr: The virtual address to read from
            size: The number of bytes to read
        """
        try:
            section = next(
                filter(lambda s: s.contains_addr(vaddr), self.cle_binary.sections)
            )
        except StopIteration as e:
            logger.error(f"Could not find section for address {vaddr}")
            raise NoSectionError(f"Could not find section for address {vaddr}") from e

        self.blob.seek(section.addr_to_offset(vaddr))
        return self.blob.read(size)

    def add_code(self, code: Code, label: Optional[str] = None) -> None:
        """Mark some code as being added on next save

        Args:
            code: The binary code being added
            label: The label to associate with the code for resolution of
                other patches referencing it
        """
        if label is None:
            label = f"code_{len(self.code_to_add)}"

        logger.info(f"Queueing code addition at label {label}: {code}")

        code.label = label

        self.code_to_add[label] = code

    def add_data(self, data: bytes, label: Optional[str] = None) -> None:
        """Add some data to the binary

        Args:
            data: The data to add
            label: The label to associate with the data location
        """
        if label is None:
            label = f"data_{len(self.data_to_add)}"

        logger.info(f"Queueing data addition at label {label}: ({len(data)} bytes)")

        self.data_to_add[label] = data

    def apply(self) -> None:
        """Apply patches to the binary"""
        # Offsets is filled in twice:
        # - first time fills in offset in the new section
        # - second time fills in offset from base address
        transform_info = TransformInfo(self.lief_binary, self.angr_project)

        transform_info.code_size = 0
        transform_info.data_size = 0
        transform_info.all_data = b""
        transform_info.data_offsets = {}
        transform_info.code_offsets = {}

        # Figure out how much space we need for code
        # Each code patch is dummy compiled and placed in a subsection that is aligned
        # to the alignment of the binary.
        for label, code in self.code_to_add.items():
            code.reset()
            code.dummy()
            compiled = code.compile(
                "dummy",
            )

            # TODO: There might be a smarter way but we don't really prioritize size
            aligned_size = self.align(len(compiled), self.alignment) * 2

            transform_info.code_offsets[label] = transform_info.code_size
            transform_info.code_size += aligned_size

        for label, data in self.data_to_add.items():
            transform_info.data_offsets[label] = len(data)
            transform_info.data_size += len(data)
            transform_info.all_data += data

        # Round up code and data sizes to the next multiple of self.alignment
        transform_info.code_size = (
            transform_info.code_size + (self.alignment - 1)
        ) & ~(self.alignment - 1)

        transform_info.data_size = (
            transform_info.data_size + (self.alignment - 1)
        ) & ~(self.alignment - 1)

        # Create new sections

        if transform_info.data_size > 0:
            new_data_segment = Segment()
            new_data_segment.content = list(transform_info.all_data)
            new_data_segment.type = SEGMENT_TYPES.LOAD
            new_data_segment.alignment = self.alignment
            new_data_segment.flags = SEGMENT_FLAGS(SEGMENT_FLAGS.R | SEGMENT_FLAGS.W)
            new_data_segment = self.lief_binary.add(new_data_segment)

            transform_info.data_base = new_data_segment.virtual_address

            # Fix up offsets to they point to the actual address
            for label, addr in transform_info.data_offsets.items():
                transform_info.data_offsets[label] = (
                    cast(int, transform_info.data_base) + addr
                )

        if transform_info.code_size > 0:
            new_code_segment = Segment()
            new_code_segment.content = list(b"\x00" * transform_info.code_size)
            new_code_segment.type = SEGMENT_TYPES.LOAD
            new_code_segment.alignment = self.alignment
            new_code_segment.flags = SEGMENT_FLAGS(SEGMENT_FLAGS.X | SEGMENT_FLAGS.R)
            new_code_segment = self.lief_binary.add(new_code_segment)

            transform_info.code_base = new_code_segment.virtual_address

            # Fix up offsets to they point to the actual address
            for label, addr in transform_info.code_offsets.items():
                transform_info.code_offsets[label] = (
                    cast(int, transform_info.code_base) + addr
                )

            # Queue writes for code
            for label, code in self.code_to_add.items():
                self.writes.append(WriteOperation(code, label))

        # Reload the angr project to get the new sections
        self.reload_blob_from_lief()
        transform_info.lief_binary = self.lief_binary
        transform_info.angr_project = self.angr_project

        for write in self.writes:
            offsets = []

            if isinstance(write.where, str):
                offsets.append(
                    transform_info.data_offsets.get(
                        write.where, transform_info.code_offsets.get(write.where, None)
                    )
                )

            elif isinstance(write.where, int):
                offsets.append(write.where)

            elif isinstance(write.where, list):
                offsets.extend(write.where)
            else:
                new_offsets = write.where(transform_info)

                if not isinstance(offsets, list):
                    offsets.append(new_offsets)
                else:
                    offsets.extend(new_offsets)

            for offset in offsets:
                logger.debug(f"Writing {write.data} to {offset:#0x}")
                transform_info.current_offset = offset

                if isinstance(write.data, bytes):
                    data = write.data

                elif isinstance(write.data, Code):

                    write.data.reset()
                    write.data.build(transform_info)
                    data = write.data.compile(
                        cast(str, write.data.label), transform_info
                    )

                    disassembly = self.angr_project.arch.disasm(data, offset)

                    logger.debug(f"Disassembly of data for label {write.data.label}:")

                    for disas_line in disassembly.splitlines():
                        logger.debug(f"  {disas_line}")
                else:
                    data = write.data(transform_info.code_offsets)

                logger.info(f"Writing {len(data)} bytes to {offset:#0x}")
                self.lief_binary.patch_address(offset, list(data))

    def save(self, where: Path) -> None:
        """Apply any pending operations and save the binary to a file

        Args:
            where: The path to the destination to save the binary
        """
        logger.info("Applying pending operations")
        self.apply()

        logger.info(f"Saving binary to {where}")
        self.lief_binary.write(str(where.resolve()))

    def asm(self, asm: str, vaddr: int) -> bytes:
        """Assemble the given assembly code at the given virtual address

        Args:
            asm: The assembly code to assemble
            vaddr: The virtual address to assemble at
        """
        return self.cle_binary.arch.asm(asm, vaddr, as_bytes=True)  # type: ignore

    def disasm(self, vaddr: int) -> CsInsn:
        """Disassemble one instruction from the binary at an address

        Args:
            vaddr: The virtual address to disassemble at
        """
        block = (
            self.binary.angr_project.kb.cfgs["CFGFast"]
            .get_any_node(vaddr, anyaddr=True)
            .block
        )

        for instr in block.capstone.insns:
            if instr.address == vaddr:
                return instr

        raise ValueError(f"Could not find instruction at {vaddr:#0x}")



"""
Code type for C code
"""

from typing import Callable, List, Optional, Union, cast

from pysquishy.squishy import Squishy
from pypatches.code.code import Code
from pypatches.transform.info import TransformInfo
from pypatches.util.libs.libs import get_lib


class CCode(Code):
    """C code container for patching

    C code should be written such that it can be compiled by
    [squishy](https://github.com/novafacing/squishy). Generally, that means that the
    `inline` keyword should not be used, code should be written as C99-like as possible,
    and that most importantly external library calls must not be used. However, you can
    write code that uses system headers using the `#include <...>` syntax, and you can
    use data structures and macros from those headers.
    """

    def compile(self, label: str, info: Optional[TransformInfo] = None) -> bytes:
        """Compile the C code to raw machine code bytes

        Args:
            label: The label to use for the code
            info: Optional transform info to use for the compilation

        Returns:
            The raw bytes of the compiled code
        """
        return cast(bytes, Squishy().compile(self.code))


def build_c_code(
    main_body: str,
    helpers: Optional[List[str]] = None,
    includes: Optional[List[str]] = None,
    extra_code: Optional[str] = None,
    dummy_transformer: Optional[
        Callable[[Union[str, bytes]], Union[str, bytes]]
    ] = None,
    build_transformer: Optional[
        Callable[[TransformInfo, Union[str, bytes]], Union[str, bytes]]
    ] = None,
    post_transformer: Optional[Callable[[TransformInfo, bytes], bytes]] = None,
) -> CCode:
    """Build a C code snippet a main body, helpers, and include files

    Generally, you should use this function instead of manually creating a
    [CCode][pypatches.code.c_code.CCode] object manually.

    Args:
        main_body: The body of the main function to build, do not include `int main(){`
            or `}`
        helpers: A list of filenames or paths to include into the program as helpers from
            `libs` like: `['liblink.c', 'libutil.c', 'libgetreg.c']. Some of these
            helpers are provided, and you can use them without a full path. The helpers
            included are:
            - `liblink.c`: A library for linking to other functions and calling library
                functions
            - `libutil.c`: A library for utility functions prefixed with `_` like
                `_strcpy` and `_strlen`
            - `libgetreg.c`: A library for getting the value of a register
            - `libsyscall.c`: A library for calling system calls
            - `libgetbase.c`: A library for getting the base address of the binary
                (note libgetbase.c requires libsyscall.c)
        includes: A list of include files like `["#include <stdint.h>", ...]`
        extra_code: Extra code to include in the program at global scope, write any
            functions you want to use in `main_body` here
        dummy_transformer: A function that is passed [self.code](#code) and returns a
            string of code to use for the dummy compilation. The result must compile but
            is not the code that will be inserted into the final patched binary, it is
            only used to obtain a size estimate. If not provided, the default is to
            return [self.code](#code) unchanged.
        build_transformer: A function that is passed
            [self.code][pypatches.code.code.Code.code] and returns a
            string of code to use for the final compilation. The result must compile and
            will be inserted into the final patched binary. If not provided, the code
            will be passed through unchanged.
        post_transformer: A function that is passed the compiled code and returns the
            final bytes to use for the patch. This is useful for doing things like
            padding the code to a certain size. If not provided, the result of
            compilation will be used unchanged.
    """
    code = ""

    if includes is not None:
        code = "\n".join(includes)
        code += "\n"

    if helpers is not None:
        for helper in helpers:
            lib_text = get_lib(helper)
            code += f"\n{lib_text}\n"

    if extra_code is not None:
        code += extra_code + "\n"

    code += """int main() {\n"""
    code += main_body + "\n"
    code += "}"

    return CCode(
        code,
        dummy_transformer=dummy_transformer,
        build_transformer=build_transformer,
        post_transformer=post_transformer,
    )



"""
Patch descriptors for various types of patches

Patches require various parameters to specify how to apply them, but some are common.

`label`, if available, allows a label to be specified. In any later patch, that label
can be used to reference the location of that patch (this is especially useful for
data patches)
"""

from dataclasses import InitVar, dataclass, field
from typing import List, Optional, Set, Union, Callable
from pypatches.code.code import Code
from pypatches.address_range import AddressRange
from pypatches.transform.info import TransformInfo


@dataclass
class NopPatch:
    """Patch that converts specific addresses or ranges of addresses to no-operations

    Attributes:
        address_ranges: The address ranges to convert to no-operations

    Args:
        addresses: Individual addresses to convert to no-operations, optional.
        address_ranges: The address ranges to convert to no-operations, optional.

    """

    addresses: InitVar[Optional[List[int]]] = None
    address_ranges: Set[AddressRange] = field(default_factory=set)

    def __post_init__(self, addresses: Optional[List[int]]) -> None:
        """Convert optionally provided addresses to address ranges

        Args:
            addresses: Addresses that can be provided in lieu of address ranges
        """
        if addresses is None:
            return

        for address in addresses:
            self.address_ranges.add(AddressRange(start=address, end=address))


@dataclass
class BranchPatch:
    """Base patch that modifies a conditional branch at a particular address

    Attributes:
        address: The address of the branch to modify
    """

    address: int


@dataclass
class InvertBranchPatch(BranchPatch):
    """Patch that inverts the true/false branch targets of a conditional branch"""


@dataclass
class AlwaysBranchPatch(BranchPatch):
    """Patch that converts a conditional branch into an unconditional branch
    that always takes the "true" branch
    """


@dataclass
class NeverBranchPatch(BranchPatch):
    """Patch that converts a conditional branch into an unconditional branch
    that always takes the "false" branch
    """


@dataclass
class SkipAndReturnPatch:
    """Patch that skips a call to a subroutine at a particular address and
    fakes a return value as if the called function had returned it using
    the default calling convention

    Attributes:
        address: The address of the call to skip
        return_value: The value to return from the call
    """

    caller_address: int
    return_value: int


@dataclass
class FunctionReplacePatch:
    """Patch that will replace a function's contents with new code

    Attributes:
        function_address: The name of the function to replace
        new_code: The new code to replace the function with
    """

    function_address: int
    new_code: Code


@dataclass
class CallerReplacePatch:
    """Patch that will redirect all or some callers of a function elsewhere

    Attributes:
        new_code: The new code to replace the function with
        function_address: The name of the function to replace
        callers: The set of callers to redirect, optional

    """

    new_code: Code
    function_address: Optional[int] = None
    callers: Set[int] = field(default_factory=set)

    def __post_init__(self) -> None:
        """
        Check that we either got a function address or callers, or both
        """
        if self.function_address is None and not self.callers:
            raise ValueError(
                "A function address and/or set of callers must be provided."
            )


@dataclass
class InitPatch:
    """Patch that will run some code immediately on entry to the program

    Attributes:
        code: The code to run
        priority: The priority of the code to run, optional
    """

    code: Code
    priority: int = 0


@dataclass
class FiniPatch:
    """Patch that will run some code upon exit from the program

    Attributes:
        code: The code to run
        priority: The priority of the code to run, optional
    """

    code: Code
    priority: int = 0


@dataclass
class DataPatch:
    """Patch that adds some data with some protections

    Attributes:
        data: The data to add
        label: The label to give the data, optional
        read: Whether the data should be readable, defaults to True
        write: Whether the data should be writable, defaults to False
        execute: Whether the data should be executable, defaults to False
    """

    data: bytes
    read: bool = True
    write: bool = False
    exec: bool = False
    label: Optional[str] = None


@dataclass
class CodePatch:
    """Base patch that holds some code

    Attributes:
        code: The code to add

    """

    code: Code


@dataclass
class AddCodePatch(CodePatch):
    """Patch that adds some code to the binary at some labeled location

    Attributes:
        label: The label to add the code at
    """

    label: Optional[str] = None


@dataclass
class ReplaceCodePatch(CodePatch):
    """Patch that replaces some code with some other code

    Attributes:
        address: The address to replace the code at, optional. This "address" can either
            be a label, an actual address, a list of addresses, a function that takes a
            [TransformInfo][TransformInfo] and returns an address, or a function that
            takes a [TransformInfo][TransformInfo] and returns a list of addresses.

    """

    address: Union[
        str,
        Callable[[TransformInfo], int],
        int,
        List[int],
        Callable[[TransformInfo], List[int]],
    ]


@dataclass
class RuntimeResolverPatch(CodePatch):
    """Patch that adds a runtime resolver to the binary"""


PatchType = Union[
    NopPatch,
    BranchPatch,
    InvertBranchPatch,
    AlwaysBranchPatch,
    NeverBranchPatch,
    SkipAndReturnPatch,
    FunctionReplacePatch,
    CallerReplacePatch,
    InitPatch,
    DataPatch,
    AddCodePatch,
    ReplaceCodePatch,
]



"""
Logging handler installer
"""


from dataclasses import dataclass
from logging import (
    CRITICAL,
    FileHandler,
    LogRecord,
    getLogger,
    getLogRecordFactory,
    setLogRecordFactory,
)
from os import getenv
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

from coloredlogs import ColoredFormatter, install


@dataclass
class LoggingConfig:
    """Logging configuration for the project"""

    default_level: str = "INFO"
    max_name_len: int = 24
    sname_intersp: str = "..."
    log_fmt: str = (
        f"{{levelname:<8s}} | "
        f"{{name:>{max_name_len + len(sname_intersp)}s}}:{{lineno:04d}} | "
        f"{{message}}"
    )
    factory: Callable[
        [
            str,
            int,
            str,
            int,
            str,
            List[Any],
            Optional[Tuple[Exception]],
            str,
            str,
            Dict[str, Any],
        ],
        LogRecord,
    ] = getLogRecordFactory()
    logdir: Optional[Path] = None
    logfile: str = "log.txt"


LOGGING_CONFIG = LoggingConfig(default_level=getenv("PATCHES_LOG_LEVEL", "DEBUG"))


def __record_factory__(
    *args: List[Any],
    **kwargs: Dict[str, Any],
) -> LogRecord:
    """
    Log record factory replacement for coloredlogging.
    Replaces `name` with a length-stabilized
    name so logs are horizontally aligned.
    :return: The new record
    :rtype: LogRecord
    """
    record = LOGGING_CONFIG.factory(*args, **kwargs)  # type: ignore
    if len(record.name) > LOGGING_CONFIG.max_name_len + len(
        LOGGING_CONFIG.sname_intersp
    ):
        record.name = (
            record.name[: LOGGING_CONFIG.max_name_len // 2]
            + LOGGING_CONFIG.sname_intersp
            + record.name[-(LOGGING_CONFIG.max_name_len // 2) :]
        )
    else:
        record.name = record.name.rjust(
            LOGGING_CONFIG.max_name_len + len(LOGGING_CONFIG.sname_intersp), " "
        )
    return record


def __init_logging__(config: LoggingConfig) -> None:
    """
    Set up logging for the project
    :param config: A logging configuration.
    """
    root_logger = getLogger()
    install(
        level=config.default_level, fmt=config.log_fmt, style="{", reconfigure=False
    )
    setLogRecordFactory(__record_factory__)
    root_logger.setLevel(config.default_level)
    root_handler = root_logger.handlers[0]
    root_handler.setFormatter(ColoredFormatter(fmt=config.log_fmt, style="{"))
    logger = getLogger(__name__)
    if config.logdir is not None:
        logfile = config.logdir / config.logfile
        assert (
            logfile.exists() and config.logdir.is_dir()
        ), f"Log directory {str(config.logdir.resolve())} does not exist!"

        fh = FileHandler(logfile)
        fh.setFormatter(ColoredFormatter(fmt=config.log_fmt, style="{"))
        root_logger.addHandler(fh)

    logger.info(f"{__name__} project logging configured.")
    parso_logger = getLogger("parso")
    parso_logger.setLevel(CRITICAL)


__init_logging__(LOGGING_CONFIG)



"""Dataclass implementing comparable address ranges.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class AddressRange:
    """An address range with a start and end.
    If start == end, this address range is just one address

    Attributes:
        start: The start address of the range
        end: The end address of the range
    """

    start: int
    end: int

    @property
    def address(self) -> bool:
        """Returns whether this range is "an address" """
        return self.start == self.end



"""Transform info
"""

from dataclasses import dataclass, field
from typing import Dict
from lief import Binary as LIEFBinary  # pylint: disable=no-name-in-module
from angr import Project


@dataclass
class TransformInfo:
    """Info for transformations of [Code][pypatches.code.code.Code] objects

    Attributes:
        lief_binary: The lief binary for the binary
        angr_project: The angr project for the binary
        code_base: The base address of the code in the binary
        code_size: The size of the code in the binary
        code_offsets: A dictionary mapping labels to offsets in the binary
        data_base: The base address of the data in the binary
        data_size: The size of the data in the binary
        data_offsets: A dictionary mapping labels to offsets in the binary
        current_offset: The offset of the current operation, usually a write
            to the binay. For example, a replacecodepatch will have at least one
            write, so this value will be set to the offset of the write before
            passing the [TransformInfo][TransformInfo] object to the build transformer
    """

    lief_binary: LIEFBinary
    angr_project: Project

    code_base: int = 0
    code_size: int = 0
    code_offsets: Dict[str, int] = field(default_factory=dict)
    data_base: int = 0
    data_size: int = 0
    data_offsets: Dict[str, int] = field(default_factory=dict)
    all_data: bytes = b""
    current_offset: int = 0



"""Accessors for various libraries to add to patches
"""

from pathlib import Path

LIBS_DIR = Path(__file__).parent


def get_lib(name: str) -> str:
    """Return the text of a library file with a given name

    Args:
        name: The name of the library file to get, either a relative path to this file
            to retrieve a built in library or a full path to a library file
            to retrieve a custom library
    """
    try:
        return (LIBS_DIR / name).read_text(encoding="utf-8")
    except FileNotFoundError:
        return Path(name).read_text(encoding="utf-8")



"""Raw machine code Code subclass"""

from typing import Optional
from pypatches.code.code import Code
from pypatches.transform.info import TransformInfo


class RawCode(Code):
    """Raw code bytes

    Raw code is a container for raw bytes of machine code. It can be used to insert
    arbitrary machine code into a binary.
    """

    def compile(self, label: str, info: Optional[TransformInfo] = None) -> bytes:
        """Return the raw bytes of the code

        Args:
            label: The label to use for the code
            info: Optional transform info to use for the compilation

        Returns:
            The raw bytes of the code
        """
        return self.code



"""Superclass for all code types
"""

from pypatches.transform.info import TransformInfo
from abc import ABC
from typing import Callable, Optional, Union

default_dummy_transformer = lambda b: b
default_build_transformer = lambda t, b: b
default_post_transformer = lambda t, b: b


class Code(ABC):
    """Superclass for other code types defining required functions

    Code objects are used to represent code that can be compiled or assembled into
    machine code. The code object is used to represent the code in a way that can be
    modified by the patching process, and then compiled or assembled into machine code
    for insertion into the binary.

    Generally, each code object goes through the following life cycle:
    - [dummy][pypatches.code.code.Code.dummy] is called to generate a dummy version of
        the code that can be used
        to estimate the size of the code in the binary.
    - [build][pypatches.code.code.Code.build] is called to generate the final version of
        the code that will be
        compiled or assembled.
    - [compile][pypatches.code.code.Code.compile] is called to compile or assemble the
        code into machine code
        bytes.
    - [post_build][pypatches.code.code.Code.post_build] is called to modify the compiled
        code after compilation
        (for example, to add a jump to the end of the code).

    Attributes:
        code (str): The code to be compiled or assembled
        original_code (str): The original code, before any modifications
        dummy_transformer (Callable[[str], str]): A function that takes this container's
            code and returns a valid compile/assemble-able code (see `self.dummy`)
        build_transformer (Callable[[TransformInfo, str], str]): A function that takes a
            `TransformInfo` object and this container's code and returns valid and final
            compile/assemble-able code.
        post_transformer (Callable[[TransformInfo, bytes], bytes]): A function that
            takes
            a `TransformInfo` object and the result of compiling this container's code
            to
            machine code bytes and returns a valid and final machine code byte string.

    Args:
        code (str): The code, to initialize with
        dummy_transformer (Callable[[str], str]): A function that takes this container's
            code and returns a valid compile/assemble-able code (see `self.dummy`),
            optional. Defaults to a function that returns the code unchanged.
        build_transformer (Callable[[TransformInfo, str], str]): A function that takes a
            `TransformInfo` object and this container's code and returns valid and final
            compile/assemble-able code, optional. Defaults to a function that returns
            the code unchanged.
        post_transformer (Callable[[TransformInfo, bytes], bytes]): A function that
            takes a `TransformInfo` object and the result of compiling this container's
            code to machine code bytes and returns a valid and final machine code byte
            string, optional. Defaults to a function that returns the compiled code
            unchanged.
    """

    code: Union[str, bytes]
    original_code: Union[str, bytes]
    label: Optional[str] = None

    def __init__(
        self,
        code: Union[str, bytes],
        dummy_transformer: Optional[
            Callable[[Union[str, bytes]], Union[str, bytes]]
        ] = None,
        build_transformer: Optional[
            Callable[[TransformInfo, Union[str, bytes]], Union[str, bytes]]
        ] = None,
        post_transformer: Optional[Callable[[TransformInfo, bytes], bytes]] = None,
    ) -> None:
        """Initialize the Code object"""
        self.code = code
        self.original_code = self.code

        self.dummy_transformer = (
            dummy_transformer
            if dummy_transformer is not None
            else default_dummy_transformer
        )

        self.build_transformer = (
            build_transformer
            if build_transformer is not None
            else default_build_transformer
        )

        self.post_transformer = (
            post_transformer
            if post_transformer is not None
            else default_post_transformer
        )

    def reset(self) -> None:
        """Reset the code to its original state without any modifications"""
        self.code = self.original_code

    def dummy(self) -> None:
        """Generate valid dummy code to determine the size required to fit this code in
        the binary.


        This function will be called to create a dummy result from the code -- the dummy
        result should be valid code (ie if this Code instance is C code it should
        compile, and if it is assembly it should assemble) but is not necessarily
        the correct code to go into the final patch.

        For example, a code snippet:

        ```
        int get_data(void) {
            return {DATA_VALUE};
        }
        ```

        might have a "dummy" result of:

        ```
        int get_data(void) {
            return 0x41414141;
        }
        ```

        The code returned by this function will never be used to produce a binary, only
        to estimate the required size for segment insertion.
        """
        self.code = self.dummy_transformer(self.code)

    def build(self, info: TransformInfo) -> None:
        """Build the code into the final version that can be built with the binary
        context into the final patch.

        Args:
            info: The `TransformInfo` object containing information about the current
                patching context.
        """
        self.code = self.build_transformer(info, self.code)

    def post_build(self, info: TransformInfo) -> bytes:
        """Modify the built code after compilation

        Args:
            info: The `TransformInfo` object containing information about the current
                patching context.

        """
        return self.post_transformer(info, self.code)

    def compile(self, label: str, info: Optional[TransformInfo] = None) -> bytes:
        """Compile or assemble the code (or do not change it if raw)

        Args:
            label: The label to use for this code
            info: The `TransformInfo` object to use for this code
        """
        raise NotImplementedError("Subclasses must implement compile()")



"""ASM subclass for assembly code used in patches
"""

from typing import Optional, cast
from pypatches.code.code import Code
from pypatches.transform.info import TransformInfo


class ASMCode(Code):
    """Assembly code container for patching

    Assembly code should be written in a format that can be assembled
    by keystone-engine. Generally, that means AT&T syntax for x86 and
    standard syntax for everything else.
    """

    def compile(self, label: str, info: Optional[TransformInfo] = None) -> bytes:
        """Assemble the assembly code to raw bytes of machine code.

        Args:
            label: The label to use for the code
            info: Optional transform info to use for the compilation

        Returns:
            The raw bytes of the compiled code
        """

        assert info is not None, "Must provide info for ASM compile"

        vaddr = info.code_offsets.get(label, 0)

        return cast(
            bytes,
            info.angr_project.arch.asm(self.code, vaddr, as_bytes=True),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypatches',
 'pypatches.code',
 'pypatches.transform',
 'pypatches.util',
 'pypatches.util.libs']

package_data = \
{'': ['*']}

install_requires = \
['angr>=9.0.0,<10.0.0',
 'archinfo>=9.0.0,<10.0.0',
 'cle>=9.0.0,<10.0.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'keystone-engine>=0.9.2,<0.10.0',
 'lief>=0.12.1,<0.13.0',
 'pysquishy==0.1.19',
 'rich>=12.4.4,<13.0.0']

setup_kwargs = {
    'name': 'pypatches',
    'version': '0.1.7',
    'description': 'Binary patching framework',
    'long_description': '# PyPatches\n\n[![Documentation Status](https://readthedocs.org/projects/binary-patches/badge/?version=main)](https://binary-patches.readthedocs.io/en/main/?badge=main)\n\n## Documentation\n\nDocumentation can be found [here](https://binary-patches.readthedocs.io/en/latest/).\n\n## Installation\n\nUp-to-date installation instructions can be found [here](docs/install.md).\n\n',
    'author': 'novafacing',
    'author_email': 'rowanbhart@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Write operation to be performed on a binary."""

from dataclasses import dataclass
from typing import Callable, List, Union, Dict

from pypatches.code.code import Code
from pypatches.transform.info import TransformInfo


@dataclass
class WriteOperation:
    """A write operation to be performed on a section

    Attributes:
        data: Either raw bytes or a function that takes the resolved
            labels for all patches and returns raw bytes
        vaddr: The virtual address or label location to write to
        where: The address to replace the code at, optional. This "address" can either
            be a label, an actual address, a list of addresses, a function that takes a
            [TransformInfo][TransformInfo] and returns an address, or a function that
            takes a [TransformInfo][TransformInfo] and returns a list of addresses.
    """

    data: Union[bytes, Callable[[Dict[str, int]], bytes], Code]
    where: Union[
        str,
        Callable[[TransformInfo], int],
        int,
        List[int],
        Callable[[TransformInfo], List[int]],
    ]
